/*
 * Copyright 2016 - 2024 Gary Talent (gary@drinkingtea.net). All rights reserved.
 */

#include <ox/claw/read.hpp>

#include <keel/media.hpp>
#include <keel/typeconv.hpp>

namespace keel {

[[nodiscard]]
static ox::Result<BaseConverter const*> findConverter(
		ox::SpanView<BaseConverter const*> const&converters,
		ox::CRStringView srcTypeName,
		int srcTypeVersion,
		ox::CRStringView dstTypeName,
		int dstTypeVersion) noexcept {
	for (auto const&c : converters) {
		if (c->matches(srcTypeName, srcTypeVersion, dstTypeName, dstTypeVersion)) {
			return c;
		}
	}
	return OxError(1, "Could not find converter");
};

static ox::Result<ox::UPtr<Wrap>> convert(
		keel::Context &ctx,
		ox::SpanView<BaseConverter const*> const&converters,
		ox::Buffer const&srcBuffer,
		ox::CRStringView srcTypeName,
		int srcTypeVersion,
		ox::CRStringView dstTypeName,
		int dstTypeVersion) noexcept {
	// look for direct converter
	auto [c, err] = findConverter(
			converters, srcTypeName, srcTypeVersion, dstTypeName, dstTypeVersion);
	if (!err) {
		return c->convertBuffToPtr(ctx, srcBuffer);
	}
	// try to chain multiple converters
	for (auto const&subConverter : converters) {
		if (!subConverter->dstMatches(dstTypeName, dstTypeVersion)) {
			continue;
		}
		const auto [intermediate, chainErr] =
			convert(ctx, converters, srcBuffer, srcTypeName, srcTypeVersion,
			        subConverter->srcTypeName(), subConverter->srcTypeVersion());
		if (!chainErr) {
			return subConverter->convertPtrToPtr(ctx, *intermediate);
		}
	}
	return OxError(1, "Could not convert between types");
}

ox::Result<ox::UPtr<Wrap>> convert(
		keel::Context &ctx,
		ox::Buffer const&srcBuffer,
		ox::CRStringView dstTypeName,
		int dstTypeVersion) noexcept {
	oxRequire(hdr, readAssetHeader(srcBuffer));
	return convert(
			ctx,
			converters(ctx),
			srcBuffer,
			hdr.clawHdr.typeName,
			hdr.clawHdr.typeVersion,
			dstTypeName,
			dstTypeVersion);
}

}
